/******************************************************************************

	romsplit.c

	Separates upper & lower ROM into high and low ROM (useful for EEPROM
	programming)
	
	Takes positive logic ROM data (as produced by a ROM readout) and
	outputs inverted ROM images.
	
	(c) 2008 by Ansgar Kueckes (kueckes@alice-dsl.net)

******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

void print_usage(char *command)
{
	printf("Usage: %s <ROM file>\n", command);
	printf("       <ROM file>  HP ROM data file to be split\n");
	return;
}


/* simulate ADA/ADB machine instruction */

void bpc_unsigned_add(unsigned long *reg, unsigned long summand, int *extend, int *overflow)
{
  long tmp,a,b;
  unsigned long result;

  /* check for overflow condition */  
  a=*reg; if (a>32767) a-=65536;
  b=summand; if (b>32767) b-=65536;
  tmp=a+b;
  if ((tmp<-32768) || (tmp>32767)) *overflow=1;

  /* integer add */
  result=(*reg)+summand;
  
  /* carry from bit 15 ? */
  if (result>65535) *extend=1;
  	
  *reg=result&0xFFFF;
}

void bpc_signed_add(long *reg, long summand, int *extend, int *overflow)
{
  long tmp,a,b,result;

  /* check for overflow condition */  
  tmp=(*reg)+summand;
  if ((tmp<-32768) || (tmp>32767)) *overflow=1;

  /* integer add */
  a=*reg; if (a<0) a+=65536; 
  b=summand; if (b<0) b+=65536;
  result=a+b;
  
  /* carry from bit 15 ? */
  if (result>65535) { result-=65536; *extend=1; }

  if (result>32767) result-=65536;
  *reg=result;
}

void bpc_rar(long *reg, int count)
{
  long a;
  int i,lsb;

  /* integer rotate right */
  a=*reg; if (a<0) a+=65536;
  for (i=0; i<count; i++) {
  	lsb=a&0x01;
  	a>>=1;
  	if (lsb) a|=0x8000;
  }

  if (a>32767) a-=65536;
  *reg=a;
}

void bpc_clear_bit15(long *reg)
{
  long a;
   
  a=*reg; if (a<0) a+=65536;
  a&=0x7FFF;
   
  //if (a>32767) a-=65536;
  *reg=a; 
}


int main(int argc, char *argv[])
{
  int count;
  char *in_file = 0;
  char hi_file[MAX_PATH];
  char lo_file[MAX_PATH];
  FILE *hfp, *lfp, *ifp;
  int blocksize=4096;
  unsigned char hi_buffer[4096];
  unsigned char lo_buffer[4096];
  unsigned char in_buffer[8192];
  size_t bytes_read;
  long checksum;
  long verify;
  int overflow, extend, excount, ovcount;
  long z;
  int i,j;

  if (argc < 2) { print_usage(argv[0]); exit(0); }

  /* command line processing */
  for ( count = 1; count < argc ; count++)
  {
	if (!strncmp(argv[count], "-", 1) && (count < argc))
		 { print_usage(argv[0]); exit(0); }

	if (!in_file) { in_file = argv[count]; continue; }
  }
  if (!in_file) { print_usage(argv[0]); exit(0); }
  	
  strcpy(hi_file, in_file);
  strcat(hi_file, "_high");
  strcpy(lo_file, in_file);
  strcat(lo_file, "_low");
  
  /* open ROM data file */
  ifp = fopen( in_file, "rb");
  if (!ifp) { fprintf(stderr, "Could not open file \"%s\".\n", in_file); exit(-1); }

  hfp = fopen( hi_file, "wb");
  if (!hfp) { fprintf(stderr, "Could not create file \"%s\".\n", hi_file); exit(-1); }
  
  lfp = fopen( lo_file, "wb");
  if (!lfp) { fprintf(stderr, "Could not create file \"%s\".\n", lo_file); exit(-1); }

  for (i=0;; i++) {
  	
    checksum=0;
    verify=0;
    overflow=0;
    excount=0;
    ovcount=0;

	/* read in one single ROM block */
	if ((bytes_read = fread((void *)in_buffer, 1, blocksize*2, ifp)) != blocksize*2) {
		if (bytes_read == 0) {
			/* done */
			exit(0);
		}
		if (ferror(ifp)) {
	   		fprintf(stderr, "Read error on file \"%s\" (%d bytes read)\n", in_file, bytes_read);
			exit(-1);
		}
	}

	/* split ROM */
	for (j=0; j<blocksize; j++) {
	  	hi_buffer[j]=~in_buffer[j*2];
	  	lo_buffer[j]=~in_buffer[j*2+1];
	  	
	  	z=in_buffer[j*2]*256+in_buffer[j*2+1];
	  	if (z>32767) z-=65536;
	  		
	  	/* update verify */ 
		extend=0;
        bpc_signed_add(&verify, z, &extend, &overflow);
        while (extend) {
        	extend=0;
        	bpc_signed_add(&verify, 1, &extend, &overflow);
        	excount++;
        }
        	
	  	/* update checksum */        	
        bpc_rar(&checksum, 3);
        if (checksum<0) {
        	//bpc_clear_bit15(&checksum);
        	bpc_rar(&checksum, 2);
        }
		bpc_signed_add(&checksum, z, &extend, &overflow);
        while (extend) {
        	extend=0;
        	bpc_signed_add(&checksum, 1, &extend, &overflow);
        	excount++;
        }
	}
    printf("block %d: verify=%ld checksum=%ld", i, verify, checksum);
	if (verify == -1) printf ("\t(ROM OK)\n"); else printf ("\t(ROM currupt)\n");
	
	if (fwrite((void *)hi_buffer, 1, blocksize, hfp) != blocksize) {
		if (ferror(hfp)) {
			fprintf(stderr, "Write error on file \"%s\" #%d\n", hi_file, errno);
			exit(-1);	
		}   
	}

	if (fwrite((void *)lo_buffer, 1, blocksize, lfp) != blocksize) {
		if (ferror(lfp)) {
			fprintf(stderr, "Write error on file \"%s\" #%d\n", lo_file, errno);
			exit(-1);	
		}   
	}
  }
  
  /* cleanup */
  if (hfp) fclose(hfp);
  if (lfp) fclose(lfp);
  if (ifp) fclose(ifp);
  exit(0);
}
